#include "x86_64_emu.h"
#include "emu_general.h"
#include "emu_defines.h"
#include "emu_misc.h"
#include "mediana.h"
#include "emu_math.h"

void emu_AAA(emu_context *ctx, struct INSTRUCTION *instr)
{
	/*
	 *  Note: This instruction incorrectly documented in Intel's materials.
	 *        The right description is:
	 *
	 *    IF (((AL and 0FH) > 9) or (AF==1)
	 *    THEN
	 *        IF CPU<286 THEN {  AL <- AL+6 }
	 *                   ELSE {  AX <- AX+6 }
	 *        AH <- AH+1
	 *        CF <- 1
	 *        AF <- 1
	 *    ELSE
	 *        CF <- 0
	 *        AF <- 0
	 *    ENDIF
	 *	AL <- AL and 0Fh
	*/	
	/* AAA affects the following flags: A,C */

	if (((AL & 0x0f) > 9) || _AF)
	{
		AX = AX + 0x106;
		setCF(); setAF();
	} else {
		clearCF(); clearAF();
	}
	AL = AL & 0x0f;
	
	/* AAA affects also the following flags: Z,S,O,P */
	/* modification of the flags is undocumented */
	/* The following behaviour seems to match the P6 and
	   its derived processors. */
	clearOF();
	clearSF(); /* sign is always 0 because bits 4-7 of AL are zeroed */
	_ZF = AL == 0;
	_PF = parity_table[AL];
}

void emu_AAD(emu_context *ctx, struct INSTRUCTION *instr)
{
	u8_t imm8, tempAL, tempAH;

	if (instr->ops[0].flags & OPERAND_FLAG_PRESENT) {
		imm8 = instr->ops[0].value.imm.imm8;
	} else {
		imm8 = 10;
	}
	tempAL = AL;
	tempAH = AH;

	AL = (tempAL + (tempAH * imm8)) & 0xFF;
	AH = 0;

	/* modification of flags A,C,O is undocumented */
	/* The following behaviour seems to match the P6 and
	   its derived processors. */
	SET_FLAGS_OSZAPC_LOGIC(AL, OPERAND_SIZE_8);
}

void emu_AAM(emu_context *ctx, struct INSTRUCTION *instr)
{
	u8_t tempAL, imm8;

	if (instr->ops[0].flags & OPERAND_FLAG_PRESENT) {
		imm8 = instr->ops[0].value.imm.imm8;
	} else {
		imm8 = 10;
	}
	if (imm8 == 0) {
		exception(EMU_EXCEPT_DE, 0);
	}
	tempAL = AL;
	AH = tempAL / imm8;
	AL = tempAL % imm8;

	/* modification of flags A,C,O is undocumented */
	/* The following behaviour seems to match the P6 and
	   its derived processors. */
	SET_FLAGS_OSZAPC_LOGIC(AL, OPERAND_SIZE_8);
}


void emu_AAS(emu_context *ctx, struct INSTRUCTION *instr)
{
	/* AAS affects the following flags: A,C */
	if (((AL & 0x0F) > 0x09) || _AF)
	{
		AX = AX - 0x106;
		setAF(); setCF();
	} else {
		clearAF(); clearCF();
	}
	AL = AL & 0x0f;

	/* AAS affects also the following flags: Z,S,O,P */
	/* modification of the flags is undocumented */
	/* The following behaviour seems to match the P6 and
	   its derived processors. */
	clearOF();
	clearSF(); /* sign is always 0 because bits 4-7 of AL are zeroed */
	_ZF = AL == 0;
	_PF = parity_table[AL];
}


void emu_DAA(emu_context *ctx, struct INSTRUCTION *instr)
{
	u8_t tmpAL = AL;
	u8_t tmpCF = 0;

	/* Validated against Intel Pentium family hardware. */
	/* DAA affects the following flags: S,Z,A,P,C       */

	if (((tmpAL & 0x0F) > 0x09) || _AF)
	{
		tmpCF = ((AL > 0xF9) || _CF);
		AL = AL + 0x06;
		setAF();
	} else {
		clearAF();
	}
	if ((tmpAL > 0x99) || _CF)
	{
		AL = AL + 0x60;
		tmpCF = 1;
	} else {
		tmpCF = 0;
	}
	clearOF(); /* undocumented flag modification */
	_SF = AL >= 0x80;
	_ZF = AL == 0;
	_CF = tmpCF;
	_PF = parity_table[AL];
}

void emu_DAS(emu_context *ctx, struct INSTRUCTION *instr)
{
	u8_t tmpAL = AL;
	u8_t tmpCF = 0;

	/* The algorithm for DAS is fashioned after the pseudo code in the
	 * Pentium Processor Family Developer's Manual, volume 3.  It seems
	 * to have changed from earlier processor's manuals.  I'm not sure
	 * if this is a correction in the algorithm printed, or Intel has
	 * changed the handling of instruction. Validated against Intel
	 * Pentium family hardware.
	*/
	/* DAS effect the following flags: A,C,S,Z,P */

	if (((tmpAL & 0x0F) > 0x09) || _AF)
	{
		tmpCF = (AL < 0x06) || _CF;
		AL = AL - 0x06;
		setAF();
	} else {
		clearAF();
	}
	if ((tmpAL > 0x99) || _CF)
	{
		AL = AL - 0x60;
		tmpCF = 1;
	}
	clearOF(); /* undocumented flag modification */
	_SF = AL >= 0x80;
	_ZF = AL == 0;
	_CF = tmpCF;
	_PF = parity_table[AL];
}

static u64_t read_CRx(emu_context *ctx, int reg)
{
	if (ctx->params->mode != EMU_MODE_16 && CPL != 0) {
		exception(EMU_EXCEPT_GP, 0);
	}
	if (reg == CREG_CODE_CR0) return ctx->regs_ctl.CR0.reg;
	if (reg == CREG_CODE_CR2) return ctx->regs_ctl.CR2.reg;
	if (reg == CREG_CODE_CR3) return ctx->regs_ctl.CR3.reg;
	if (reg == CREG_CODE_CR4) return ctx->regs_ctl.CR4.reg;
	exception(EMU_EXCEPT_UD, 0);
}

static u64_t read_DRx(emu_context *ctx, int reg)
{
	//FIXME
	exception(EMU_EXCEPT_UD, 0);
}

static void writeSEG(emu_context *ctx, int reg, u16_t data)
{
	if ( (ctx->params->mode != EMU_MODE_16) && 
		 (get_regseg(reg) != data) && (ctx->params->do_loadsreg != NULL) )
	{
		if (ctx->params->do_loadsreg(ctx, REG_TYPE_SEG, reg, data) == EMU_STATUS_OK) {
			ctx->regs_seg.regs[reg] = data;
		}
	} else {
		ctx->regs_seg.regs[reg] = data;
	}
}

static void writeCRx(emu_context *ctx, int reg, u64_t data)
{
	//FIXME
	exception(EMU_EXCEPT_UD, 0);
}

static void writeDRx(emu_context *ctx, int reg, u64_t data)
{
	//FIXME
	exception(EMU_EXCEPT_UD, 0);
}

void emu_MOV(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t tmp;

	if ( (instr->ops[1].flags & OPERAND_TYPE_REG) && (instr->ops[1].value.reg.type != REG_TYPE_GEN) )
	{
		switch (instr->ops[1].value.reg.type) {
			case REG_TYPE_SEG: tmp = get_regseg(OP1_REG_CODE);    break;
			case REG_TYPE_CR:  tmp = read_CRx(ctx, OP1_REG_CODE); break;
			case REG_TYPE_DBG: tmp = read_DRx(ctx, OP1_REG_CODE); break;
			default: exception(EMU_EXCEPT_UD, 0);
		}
	} else {
		tmp = load_operand(1, 0);
	}	
	if ( (instr->ops[0].flags & OPERAND_TYPE_REG) && (instr->ops[0].value.reg.type != REG_TYPE_GEN) ) 
	{
		switch (instr->ops[0].value.reg.type) {
			case REG_TYPE_SEG: writeSEG(ctx, OP0_REG_CODE, u16t(tmp)); break;
			case REG_TYPE_CR:  writeCRx(ctx, OP0_REG_CODE, u64t(tmp)); break;
			case REG_TYPE_DBG: writeDRx(ctx, OP0_REG_CODE, u64t(tmp)); break;
			default: exception(EMU_EXCEPT_UD, 0);
		}
	} else {
		save_operand(0, tmp, 0);
	}
}

void emu_LEA(emu_context *ctx, struct INSTRUCTION *instr)
{
	if ( (instr->ops[0].flags & OPERAND_TYPE_REG) == 0 || 
		 (instr->ops[1].flags & OPERAND_TYPE_MEM) == 0 ) 
	{
		exception(EMU_EXCEPT_UD, 0);
	}
	save_operand(0, get_address(1, NULL), 0);
}

void emu_ADD(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);
	u64_t sum = op0 + op1;
	
	save_operand(0, sum, 0);
	SET_FLAGS_OSZAPC_ADD(op0, op1, sum, OP0_SIZE);
}

void emu_ADC(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);
	u64_t sum = op0 + op1 + _CF;
	
	save_operand(0, sum, 0);
	SET_FLAGS_OSZAPC_ADC(op0, op1, sum, OP0_SIZE);
}

void emu_SUB(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);
	u64_t dif = op0 - op1;

	save_operand(0, dif, 0);
	SET_FLAGS_OSZAPC_SUB(op0, op1, dif, OP0_SIZE);
}

void emu_SBB(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);
	u64_t dif = op0 - op1 - _CF;

	save_operand(0, dif, 0);
	SET_FLAGS_OSZAPC_SBB(op0, op1, dif, OP0_SIZE);
}

void emu_AND(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);
	u64_t res = op0 & op1;

	save_operand(0, res, 0);
	SET_FLAGS_OSZAPC_LOGIC(res, OP0_SIZE);
}

void emu_INC(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t sum = op0 + 1;

	save_operand(0, sum, 0);
	SET_FLAGS_OSZAPC_INC(sum, OP0_SIZE);
}

void emu_DEC(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t dif = op0 - 1;

	save_operand(0, dif, 0);
	SET_FLAGS_OSZAPC_DEC(dif, OP0_SIZE);
}

void emu_CLC(emu_context *ctx, struct INSTRUCTION *instr)
{
	clearCF();
}

void emu_STC(emu_context *ctx, struct INSTRUCTION *instr)
{
	setCF();
}

void emu_CLD(emu_context *ctx, struct INSTRUCTION *instr)
{
	clearDF();
}

void emu_STD(emu_context *ctx, struct INSTRUCTION *instr)
{
	setDF();
}

void emu_CLI(emu_context *ctx, struct INSTRUCTION *instr)
{
	if (ctx->params->mode != EMU_MODE_16 && IOPL < CPL) {
		exception(EMU_EXCEPT_GP, 0);
	}
	clearIF();
}

void emu_STI(emu_context *ctx, struct INSTRUCTION *instr)
{
	if (ctx->params->mode != EMU_MODE_16 && IOPL < CPL) {
		exception(EMU_EXCEPT_GP, 0);
	}
	setIF();
}

void emu_BSF(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op1 = load_operand(1, 0);
	u64_t op0;
	
	if (op1 != 0) {
		for (op0 = 0; (op1 & (1LL << op0)) == 0; op0++);
		clearZF();
		save_operand(0, op0, 0);		
	} else {
		setZF();
	}
}

void emu_BSR(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op1 = load_operand(1, 0);
	u64_t op0;

	if (op1 != 0) {
		for (op0 = 63; (op1 & (1LL << op0)) == 0; op0--);
		clearZF();
		save_operand(0, op0, 0);		
	} else {
		setZF();
	}
}

void emu_BT(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op1 = load_operand(1, 0);
	u64_t bit = bit_op_bit(op1);
	u64_t off = bit_op_off(op1);
	u64_t op0 = load_operand(0, off);
	
	_CF = (op0 >> bit) & 1;
}

void emu_BTC(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op1 = load_operand(1, 0);
	u64_t bit = bit_op_bit(op1);
	u64_t off = bit_op_off(op1);
	u64_t op0 = load_operand(0, off);

	_CF = (op0 >> bit) & 1;
	save_operand(0, op0 ^ (1ULL << bit), off);
}

void emu_BTR(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op1 = load_operand(1, 0);
	u64_t bit = bit_op_bit(op1);
	u64_t off = bit_op_off(op1);
	u64_t op0 = load_operand(0, off);

	_CF = (op0 >> bit) & 1;
	save_operand(0, op0 & ~(1ULL << bit), off);
}

void emu_BTS(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op1 = load_operand(1, 0);
	u64_t bit = bit_op_bit(op1);
	u64_t off = bit_op_off(op1);
	u64_t op0 = load_operand(0, off);

	_CF = (op0 >> bit) & 1;
	save_operand(0, op0 | (1ULL << bit), off);
}

void emu_BSWAP(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op = get_reggen64(OP0_REG_CODE);

	switch (OP0_SIZE) {
		case OPERAND_SIZE_16: set_reggen16(OP0_REG_CODE, 0); break;
		case OPERAND_SIZE_32: op = BSWAP_32(op); set_reggen32(OP0_REG_CODE, op); break;
		case OPERAND_SIZE_64: op = BSWAP_64(op); set_reggen64(OP0_REG_CODE, op); break;
	}
}

/* CBW / CWDE / CDQE */
void emu_CBW(emu_context *ctx, struct INSTRUCTION *instr)
{
	switch (instr->opsize) {
		case OPERAND_SIZE_16: AX  = S_EXTD(AL,  8,  16); break;
		case OPERAND_SIZE_32: EAX = S_EXTD(AX,  16, 32); break;
		case OPERAND_SIZE_64: RAX = S_EXTD(EAX, 32, 64); break;
	}	
}

void emu_CLTS(emu_context *ctx, struct INSTRUCTION *instr)
{
	if (ctx->params->mode != EMU_MODE_16 && CPL != 0) {
		exception(EMU_EXCEPT_GP, 0);
	}
	ctx->regs_ctl.CR0.TS = 0;
}

void emu_CMC(emu_context *ctx, struct INSTRUCTION *instr)
{
	_CF = !_CF;
}

void emu_CMOVcc(emu_context *ctx, struct INSTRUCTION *instr, int if_ok)
{
	u64_t op = if_ok ? load_operand(1, 0) : get_reggen64(OP0_REG_CODE);

	if (OP0_SIZE == OPERAND_SIZE_32) {
		op &= 0xFFFFFFFF;
	}
	set_reggen64(OP0_REG_CODE, op);
}

void emu_ARPL(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);

	if ( (op0 & 3) < (op1 & 3) )
	{
		op0 = (op0 & 0xFFFC) | (op1 & 3);

		save_operand(0, op0, 0);
		setZF();
	} else {
		clearZF();
	}
}

void emu_BOUND(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t test = load_operand(0, 0);
	int   sreg;
	u64_t offs = get_address(0, &sreg);
	u32_t bmin, bmax;

	if (OP0_SIZE == OPERAND_SIZE_16) {
		bmin = read_mem16(sreg, offs); bmax = read_mem16(sreg, offs + 2);
	} else {
		bmin = read_mem32(sreg, offs); bmax = read_mem32(sreg, offs + 4);
	}
	if (test < bmin || test > bmax) {
		exception(EMU_EXCEPT_BR, 0);
	}
}

void emu_CMP(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);
	u64_t dif = op0 - op1;

	SET_FLAGS_OSZAPC_SUB(op0, op1, dif, OP0_SIZE);
}


void emu_CMPXCHG(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t acc = get_reggen(REG_CODE_AX, OP0_SIZE);
	u64_t op0 = load_operand(0, 0);
	u64_t dif = acc - op0;

	SET_FLAGS_OSZAPC_SUB(acc, op0, dif, OP0_SIZE);

	if (dif == 0) {
		save_operand(0, load_operand(1, 0), 0);
	} else {
		set_reggen(REG_CODE_AX, acc, OP0_SIZE);
	}
}


void emu_CMPXCHG8B(emu_context *ctx, struct INSTRUCTION *instr)
{
	int   sreg;
	u64_t offs = get_address(0, &sreg);
	u64_t ophi, oplo;

	if (OP0_SIZE == OPERAND_SIZE_128)
	{
		if (offs & 0xF) {
			exception(EMU_EXCEPT_GP, 0);
		}
		oplo = read_mem64(sreg, offs + 0); 
		ophi = read_mem64(sreg, offs + 8);

		if (RAX == oplo && RDX == ophi)
		{
			write_mem64(sreg, offs + 0, RBX);
			write_mem64(sreg, offs + 8, RCX);
			setZF();			
		} else {
			RAX = oplo; RDX = ophi;
			clearZF();
		}
	} else
	{
		oplo = read_mem64(sreg, offs);
		
		if (oplo == ((u64t(EDX) << 32) | EAX)) {
			write_mem64(sreg, offs, (u64t(ECX) << 32) | EBX);
			setZF();
		} else {
			EDX = u32t(oplo >> 32); EAX = u32t(oplo);
			clearZF();
		}
	}		
}

/* CWQ / CDQ / CQO */
void emu_CWD(emu_context *ctx, struct INSTRUCTION *instr)
{
	switch (instr->opsize) {
		case OPERAND_SIZE_16: DX  = AX  & LAST_BIT(16) ? BIT_MASK(16) : 0; break;
		case OPERAND_SIZE_32: EDX = EAX & LAST_BIT(32) ? BIT_MASK(32) : 0; break;
		case OPERAND_SIZE_64: RDX = RAX & LAST_BIT(64) ? BIT_MASK(64) : 0; break;
	}
}


void emu_DIV(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op1 = load_operand(0, 0);
	
	if (op1 == 0) {
		exception(EMU_EXCEPT_DE, 0);
	}
	switch (OP0_SIZE)
	{
		case OPERAND_SIZE_8:  EMU_DO_DIV(op1, AH,  AL,  u16_t, 8); break;
		case OPERAND_SIZE_16: EMU_DO_DIV(op1, DX,  AX,  u32_t, 16); break;
		case OPERAND_SIZE_32: EMU_DO_DIV(op1, EDX, EAX, u64_t, 32); break;
		case OPERAND_SIZE_64:
			{
				u128_t op0_128, quotient;
				u64_t  remainder;

				op0_128.lo = RAX;
				op0_128.hi = RDX;

				long_div(&quotient, &remainder, &op0_128, op1);
				
				if (quotient.hi != 0) {
					exception(EMU_EXCEPT_DE, 0);
				}
				RAX = quotient.lo;
				RDX = remainder; 
			}
		break;
	}
}


void emu_IDIV(emu_context *ctx, struct INSTRUCTION *instr)
{
	s64_t op1 = load_operand(0, 0);
	
	if (op1 == 0) {
		exception(EMU_EXCEPT_DE, 0);
	}	
	switch (OP0_SIZE)
	{
		case OPERAND_SIZE_8:  EMU_DO_IDIV(op1, AH,  AL,  s8_t,  s16_t, 8);  break;
		case OPERAND_SIZE_16: EMU_DO_IDIV(op1, DX,  AX,  s16_t, s32_t, 16); break;
		case OPERAND_SIZE_32: EMU_DO_IDIV(op1, EDX, EAX, s32_t, s64_t, 32); break;
		case OPERAND_SIZE_64:
			{
				s128_t op0_128, quotient;
				s64_t  remainder;
				
				op0_128.lo = RAX;
				op0_128.hi = RDX;

				long_idiv(&quotient, &remainder, &op0_128, op1);
				
				if ((!(quotient.lo & LAST_BIT(64)) && quotient.hi != 0) ||
				      (quotient.lo & LAST_BIT(64)) && quotient.hi != BIT_MASK(64))
				{
					exception(EMU_EXCEPT_DE, 0);
				}
				RAX = quotient.lo;
				RDX = remainder; 
			}
		break;
	}
}

void emu_MUL(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op1 = load_operand(0, 0);

	switch (OP0_SIZE)
	{
		case OPERAND_SIZE_8:  ENU_DO_MUL(AL,  AH,  op1, 8,  16); break;
		case OPERAND_SIZE_16: ENU_DO_MUL(AX,  DX,  op1, 16, 32); break;
		case OPERAND_SIZE_32: ENU_DO_MUL(EAX, EDX, op1, 32, 64); break;
		case OPERAND_SIZE_64: 
			{
				u128_t res;
								
				long_mul(&res, RAX, op1);

				RAX = res.lo;
				RDX = res.hi;

				SET_FLAGS_OSZAPC_LOGIC(res.lo, OPERAND_SIZE_64);

				if (res.hi != 0) {
					setOF(); setCF();
				}
			}
		break;
	}
}

void emu_IMUL(emu_context *ctx, struct INSTRUCTION *instr)
{
	s64_t  op0, op1;
	s128_t res;
	
	if (instr->ops[1].flags & OPERAND_FLAG_PRESENT)
	{
		if (instr->ops[2].flags & OPERAND_FLAG_PRESENT) {		
			op1 = load_operand(2, 0);
		} else {
			op1 = load_operand(0, 0);			
		}
		op0 = load_operand(1, 0);
	} else {
		op0 = get_reggen(REG_CODE_AX, OP0_SIZE);
		op1 = load_operand(0, 0);
	}
	switch (OP0_SIZE)
	{
		case OPERAND_SIZE_8:  EMU_DO_IMUL(res.lo, op0, op1, 8,  16); break;
		case OPERAND_SIZE_16: EMU_DO_IMUL(res.lo, op0, op1, 16, 32); break;
		case OPERAND_SIZE_32: EMU_DO_IMUL(res.lo, op0, op1, 32, 64); break;
		case OPERAND_SIZE_64:
			long_imul(&res, op0, op1);

			SET_FLAGS_OSZAPC_LOGIC(res.lo, OPERAND_SIZE_64);

			if (u64t(res.hi) + (res.lo >> 63) != 0) {
				setOF(); setCF();
			}
		break;
	}
	if (instr->ops[1].flags & OPERAND_FLAG_PRESENT) {
		save_operand(0, res.lo, 0);
	} else 
	{
		switch (OP0_SIZE) {
			case OPERAND_SIZE_8:  AX  = u16t(res.lo); break;
			case OPERAND_SIZE_16: DX  = u16t(res.lo >> 16); AX  = u16t(res.lo); break;
			case OPERAND_SIZE_32: EDX = u32t(res.lo >> 32); EAX = u32t(res.lo); break;
			case OPERAND_SIZE_64: RDX = res.hi; RAX = res.lo; break;
		}
	}
}

static void pushSEG(emu_context *ctx, int sreg)
{
	u16_t seg = get_regseg(sreg);

	switch (ctx->params->mode) {
		case EMU_MODE_16: push16(seg); break;
		case EMU_MODE_32: push32(seg); break;
		case EMU_MODE_64: push64(seg); break;
	}
}

static void popSEG(emu_context *ctx, int sreg)
{
	u16_t seg;

	switch (ctx->params->mode) {
		case EMU_MODE_16: seg = u16t(pop16()); break;
		case EMU_MODE_32: seg = u16t(pop32()); break;
		case EMU_MODE_64: seg = u16t(pop64()); break;
	}
	writeSEG(ctx, sreg, seg);
}

void emu_PUSH(emu_context *ctx, struct INSTRUCTION *instr)
{
	if ( (instr->ops[0].flags & OPERAND_TYPE_REG) && (instr->ops[0].value.reg.type == REG_TYPE_SEG) ) {
		pushSEG(ctx, OP0_REG_CODE);
	} else {
		emu_push(ctx, load_operand(0, 0), OP0_SIZE);
	}
}

void emu_PUSHA(emu_context *ctx, struct INSTRUCTION *instr)
{
	u32_t tmp = ESP;

	if (instr->opsize == OPERAND_SIZE_32) {
		push32(EAX); push32(ECX); push32(EDX); push32(EBX);
		push32(tmp); push32(EBP); push32(ESI); push32(EDI);
	} else {
		push16(AX);  push16(CX); push16(DX); push16(BX);
		push16(tmp); push16(BP); push16(SI); push16(DI);
	}
}

void emu_PUSHF(emu_context *ctx, struct INSTRUCTION *instr)
{
	switch (instr->opsize) {
		case OPERAND_SIZE_16: push16(EFL & 0x00FFFF); break;
		case OPERAND_SIZE_32: push32(EFL & 0xFCFFFF); break;
		case OPERAND_SIZE_64: push64(RFL & 0xFCFFFF); break;
	}
}

void emu_POP(emu_context *ctx, struct INSTRUCTION *instr)
{
	if ( (instr->ops[0].flags & OPERAND_TYPE_REG) && (instr->ops[0].value.reg.type == REG_TYPE_SEG) ) {
		popSEG(ctx, OP0_REG_CODE);
	} else {
		save_operand(0, emu_pop(ctx, OP0_SIZE), 0);
	}
}

void emu_POPA(emu_context *ctx, struct INSTRUCTION *instr)
{
	if (instr->opsize == OPERAND_SIZE_32) {
		EDI = pop32(); ESI = pop32(); EBP = pop32(); pop32();
		EBX = pop32(); EDX = pop32(); ECX = pop32(); EAX = pop32();
	} else {
		DI = pop16(); SI = pop16(); BP = pop16(); pop16();
		BX = pop16(); DX = pop16(); CX = pop16(); AX = pop16();
	}
}

void emu_POPF(emu_context *ctx, struct INSTRUCTION *instr)
{
	u32_t flags = u32t(emu_pop(ctx, instr->opsize));
	u32_t mask  = FLAGS_OSZAPC_MASK | FLAGS_TF_MASK | FLAGS_DF_MASK | FLAGS_NT_MASK;

	if (ctx->params->mode != EMU_MODE_16)
	{
		if (CPL == 0) {
			mask |= FLAGS_IOPL_MASK;
		}
		if (CPL <= IOPL) {
			mask |= FLAGS_IF_MASK;
		}
	} else {
		mask |= FLAGS_IOPL_MASK | FLAGS_IF_MASK;
	}
	if (instr->opsize == EMU_MODE_32) {
		mask |= FLAGS_RF_MASK;
	}
	if (instr->opsize == EMU_MODE_64) {
		mask |= FLAGS_RF_MASK | FLAGS_AC_MASK | FLAGS_ID_MASK; 
	}
	EFL = (EFL & ~mask) | (flags & mask);
}

void emu_JMP(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t dest;

	if (instr->ops[0].flags & OPERAND_TYPE_IMM) {
		dest = load_imm_op(0) + RIP + instr->length;
	} else {
		dest = load_operand(0, 0);
	}
	if (ctx->params->mode == EMU_MODE_64) {
		RIP = u64t(dest);
	} else {
		EIP = u32t(dest);
	}
	ctx->jumped = 1;
}

void emu_JCXZ(emu_context *ctx, struct INSTRUCTION *instr)
{
	if (get_reggen(REG_CODE_CX, instr->opsize) == 0) {
		emu_JMP(ctx, instr);
	}
}

void emu_JMPF(emu_context *ctx, struct INSTRUCTION *instr)
{
	u16_t seg;
	u32_t dest;

	if (instr->ops[0].flags & OPERAND_TYPE_DIR) 
	{
		if (ctx->params->mode == EMU_MODE_16) {
			seg  = instr->ops[0].value.far_addr.far_addr32.seg;
			dest = instr->ops[0].value.far_addr.far_addr32.offset;
		} else {
			seg  = instr->ops[0].value.far_addr.far_addr48.seg;
			dest = instr->ops[0].value.far_addr.far_addr48.offset;
		}
	} else
	{
		int   sreg;
		u64_t offs = get_address(0, &sreg);

		if (ctx->params->mode == EMU_MODE_16) {
			dest = read_mem16(sreg, offs); seg = read_mem16(sreg, offs + 2);
		} else {
			dest = read_mem32(sreg, offs); seg = read_mem16(sreg, offs + 4);
		}
	}
	writeSEG(ctx, SREG_CODE_CS, seg); EIP = dest; ctx->jumped = 1;
}

void emu_CALL(emu_context *ctx, struct INSTRUCTION *instr)
{
	switch (ctx->params->mode) {
		case EMU_MODE_16: push16(EIP + instr->length); break;
		case EMU_MODE_32: push32(EIP + instr->length); break;
		case EMU_MODE_64: push64(RIP + instr->length); break;
	}
	emu_JMP(ctx, instr);
}

void emu_CALLF(emu_context *ctx, struct INSTRUCTION *instr)
{
	pushSEG(ctx, SREG_CODE_CS);

	if (ctx->params->mode == EMU_MODE_16) {
		push16(EIP + instr->length);
	} else {
		push32(EIP + instr->length);
	}
	emu_JMPF(ctx, instr);
}


void emu_SAHF(emu_context *ctx, struct INSTRUCTION *instr)
{
	_SF = (AH & 0x80) != 0;
	_ZF = (AH & 0x40) != 0;
	_AF = (AH & 0x10) != 0;
	_CF = (AH & 0x01) != 0;
	_PF = (AH & 0x04) != 0;
}

void emu_LAHF(emu_context *ctx, struct INSTRUCTION *instr)
{
	AH = EFL & 0xFF;
}

void emu_SALC(emu_context *ctx, struct INSTRUCTION *instr)
{
	AH = (_CF != 0) ? 0xFF : 0;
}

void emu_OR(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);
	u64_t res = op0 | op1;

	save_operand(0, res, 0);
	SET_FLAGS_OSZAPC_LOGIC(res, OP0_SIZE);
}

void emu_NOT(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t res = ~op0;

	save_operand(0, res, 0);
	SET_FLAGS_OSZAPC_LOGIC(res, OP0_SIZE);
}

void emu_XOR(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);
	u64_t res = op0 ^ op1;

	save_operand(0, res, 0);
	SET_FLAGS_OSZAPC_LOGIC(res, OP0_SIZE);
}

void emu_NEG(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	s64_t res;

	switch (OP0_SIZE) {
		case OPERAND_SIZE_8:  res = -s8t(op0);  break;
		case OPERAND_SIZE_16: res = -s16t(op0); break;
		case OPERAND_SIZE_32: res = -s32t(op0); break;
		case OPERAND_SIZE_64: res = -s64t(op0); break;
	}
	save_operand(0, res, 0);
	SET_FLAGS_OSZAPC_NEG(res, OP0_SIZE);
}


void emu_ROL(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t res, op0 = load_operand(0, 0);
	u32_t shift;

	if (shift = load_operand(1, 0) & (OP0_BITS - 1))
	{
		res = (op0 << shift) | (op0 >> (OP0_BITS - shift));
		save_operand(0, res, 0);
		
		/* CF = first_bit 
		   OF = last_bit ^ first_bit */
		_CF = ((res >> 0) & 1);
		_OF = ((res >> (OP0_BITS - 1)) & 1) ^ _CF;
	}
}

void emu_ROR(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t res, op0 = load_operand(0, 0);
	u32_t shift;
	
	if (shift = load_operand(1, 0) & (OP0_BITS - 1))
	{
		res = (op0 >> shift) | (op0 << (OP0_BITS - shift));
		save_operand(0, res, 0);

		/* CF = last_bit 
		   OF = prev_last_bit ^ last_bit */
		_CF = ((res >> (OP0_BITS - 1)) & 1);
		_OF = ((res >> (OP0_BITS - 2)) & 1) ^ _CF;
	}
}

void emu_RCL(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t res, op0 = load_operand(0, 0);
	u32_t shift;
	
	if (shift = (load_operand(1, 0) & ROT_MASK) % (OP0_BITS + 1))
	{
		res = (op0 << shift) | 
			  (_CF << (shift - 1)) | (op0 >> ((OP0_BITS + 1) - shift));
		save_operand(0, res, 0);

		/* CF = bit_(max_bits - shift)
		   OF = last_bit ^ bit_(max_bits - shift) */
		_CF = ((res >> (OP0_BITS - shift)) & 1);
		_OF = ((res >> (OP0_BITS - 1)) & 1) ^ _CF;
	}
}

void emu_RCR(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t res, op0 = load_operand(0, 0);
	u32_t shift;
	
	if (shift = (load_operand(1, 0) & ROT_MASK) % (OP0_BITS + 1))
	{
		res = (op0 >> shift) | 
			  (_CF << (OP0_BITS - shift)) | (op0 << ((OP0_BITS + 1) - shift));
		save_operand(0, res, 0);

		/* CF = bit_(max_bits - shift)
		   OF = last_bit ^ bit_(max_bits - shift) */
		_CF = ((res >> (OP0_BITS - shift)) & 1);
		_OF = ((res >> (OP0_BITS - 1)) & 1) ^ _CF;
	}
}

void emu_SAR(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t res, op0 = load_operand(0, 0);
	u32_t shift;
	u16_t _cf;

	if (shift = load_operand(1, 0) & ROT_MASK)
	{
		if (shift < OP0_BITS) 
		{
			res = (op0 >> shift) | 
				  (op0 & sign_bits[OP0_SIZE] ? (0xFFFFFFFFFFFFFFFF << (OP0_BITS - shift)) : 0);			
			_cf = (op0 >> (shift - 1)) & 1;
		} else {
			res = op0 & sign_bits[OP0_SIZE] ? 0xFFFFFFFFFFFFFFFF : 0;
			_cf = res & 1;
		}
		SET_FLAGS_OSZAPC_LOGIC(res, OP0_SIZE);
		_CF = _cf;		
		save_operand(0, res, 0);
	}
}

void emu_SHL(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t res, op0 = load_operand(0, 0);
	u32_t shift;
	
	if (shift = load_operand(1, 0) & ROT_MASK)
	{
		res = shift <= OP0_BITS ? op0 << shift : 0;		
		save_operand(0, res, 0);

		SET_FLAGS_OSZAPC_LOGIC(res, OP0_SIZE);
		_CF = ((op0 >> (OP0_BITS - shift)) & 1); 
		_OF = ((res >> (OP0_BITS - 1)) & 1) ^ _CF;
	}
}

void emu_SHR(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t res, op0 = load_operand(0, 0);
	u32_t shift;
	
	if (shift = load_operand(1, 0) & ROT_MASK)
	{
		res = op0 >> shift; save_operand(0, res, 0);

		SET_FLAGS_OSZAPC_LOGIC(res, OP0_SIZE);
		_CF = (op0 >> (shift - 1)) & 1; 
		_OF = (((res << 1) ^ res) >> (OP0_BITS - 1)) & 1;
	}
}

void emu_SHLD(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);
	u32_t shift;
	
	if (shift = load_operand(2, 0) & ROT_MASK)
	{
		u64_t res = (op0 << shift) | (op1 >> (OP0_BITS - shift));
		save_operand(0, res, 0);
		
		SET_FLAGS_OSZAPC_LOGIC(res, OP0_SIZE);
		_CF = ((op0 >> (OP0_BITS - shift)) & 1);
		_OF = ((res >> (OP0_BITS - 1)) & 1) ^ _CF;
	}
}

void emu_SHRD(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);
	u32_t shift;
	
	if (shift = load_operand(2, 0) & ROT_MASK)
	{
		u64_t res = (op1 << (OP0_BITS - shift)) | (op0 >> shift);
		save_operand(0, res, 0);
		
		SET_FLAGS_OSZAPC_LOGIC(res, OP0_SIZE);

		_CF = ((op0 >> (shift - 1)) & 1);
		_OF = (((res << 1) ^ res) >> (OP0_BITS - 1)) & 1;
	}
}

void emu_TEST(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t res = load_operand(0, 0) & 
		        load_operand(1, 0);

	SET_FLAGS_OSZAPC_LOGIC(res, OP0_SIZE);
}

void emu_MOVSX(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op1 = load_operand(1, 0);
	s64_t res = OP1_SIZE == OPERAND_SIZE_8 ? s8t(op1) : s16t(op1);

	save_operand(0, res, 0);
}

void emu_MOVSXD(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op1 = load_operand(1, 0);
	s64_t res = s32t(op1);

	save_operand(0, res, 0);
}

void emu_MOVZX(emu_context *ctx, struct INSTRUCTION *instr)
{
	save_operand(0, load_operand(1, 0), 0);
}

static void emu_repeat(emu_context *ctx, struct INSTRUCTION *instr, p_emufun func)
{
	u64_t cx;

	ctx->in_repeat++;

	while (cx = get_reg_addr(REG_CODE_CX)) {
		func(ctx, instr); ctx->ticks++;
		set_reg_addr(REG_CODE_CX, cx - 1);	
	}
	ctx->in_repeat--;
	/* zero extension of RCX */
	if (ctx->params->mode == EMU_MODE_32) RCX = ECX;
}

static void emu_repeat_Z(emu_context *ctx, struct INSTRUCTION *instr, p_emufun func)
{
	u64_t cx;

	ctx->in_repeat++;

	while (cx = get_reg_addr(REG_CODE_CX)) {
		func(ctx, instr); ctx->ticks++;
		set_reg_addr(REG_CODE_CX, cx - 1);
		if (_ZF == 0) break;
	}
	ctx->in_repeat--;
	/* zero extension of RCX */
	if (ctx->params->mode == EMU_MODE_32) RCX = ECX;
}

static void emu_repeatNZ(emu_context *ctx, struct INSTRUCTION *instr, p_emufun func)
{
	u64_t cx;

	ctx->in_repeat++;

	while (cx = get_reg_addr(REG_CODE_CX)) {
		func(ctx, instr); ctx->ticks++;
		set_reg_addr(REG_CODE_CX, cx - 1);
		if (_ZF != 0) break;
	}
	ctx->in_repeat--;
	/* zero extension of RCX */
	if (ctx->params->mode == EMU_MODE_32) RCX = ECX;
}

void emu_LODS(emu_context *ctx, struct INSTRUCTION *instr)
{
	if ( (instr->prefixes & INSTR_PREFIX_REP_MASK) && (ctx->in_repeat == 0) ) {
		emu_repeat(ctx, instr, &emu_LODS);
	} else
	{
		u64_t si = get_reg_addr(REG_CODE_SI);
		u64_t op = read_mem_size(SREG_CODE_DS, si, instr->opsize);

		set_reggen(REG_CODE_AX, op, instr->opsize);

		if (_DF != 0) {
			si -= instr->opsize;
		} else {
			si += instr->opsize;
		}
		set_reg_addr(REG_CODE_SI, si);

		/* zero extension of RSI */
		if (ctx->params->mode == EMU_MODE_32) RSI = ESI;
	}
}

void emu_STOS(emu_context *ctx, struct INSTRUCTION *instr)
{
	if ( (instr->prefixes & INSTR_PREFIX_REP_MASK) && (ctx->in_repeat == 0) ) {
		emu_repeat(ctx, instr, &emu_STOS);
	} else
	{
		u64_t di = get_reg_addr(REG_CODE_DI);
		u64_t op = get_reggen(REG_CODE_AX, instr->opsize);

		write_mem_size(SREG_CODE_ES, di, op, instr->opsize);

		if (_DF != 0) {
			di -= instr->opsize;
		} else {
			di += instr->opsize;
		}
		set_reg_addr(REG_CODE_DI, di);

		/* zero extension of RDI */
		if (ctx->params->mode == EMU_MODE_32) RDI = EDI;
	}
}

void emu_MOVS(emu_context *ctx, struct INSTRUCTION *instr)
{
	if ( (instr->prefixes & INSTR_PREFIX_REP_MASK) && (ctx->in_repeat == 0) ) {
		emu_repeat(ctx, instr, &emu_MOVS);
	} else 
	{
		u64_t si = get_reg_addr(REG_CODE_SI);
		u64_t di = get_reg_addr(REG_CODE_DI);
		u64_t op = read_mem_size(SREG_CODE_DS, si, instr->opsize);

		write_mem_size(SREG_CODE_ES, di, op, instr->opsize);

		if (_DF != 0) {
			si -= instr->opsize; di -= instr->opsize;
		} else {
			si += instr->opsize; di += instr->opsize;
		}
		set_reg_addr(REG_CODE_SI, si);
		set_reg_addr(REG_CODE_DI, di);

		/* zero extension of RSI/RDI */
		if (ctx->params->mode == EMU_MODE_32) {
			RSI = ESI; RDI = EDI;
		}
	}
}

void emu_SCAS(emu_context *ctx, struct INSTRUCTION *instr)
{
	if ( (instr->prefixes & INSTR_PREFIX_REP_MASK) && (ctx->in_repeat == 0) ) 
	{
		if (instr->prefixes & INSTR_PREFIX_REPZ) {
			emu_repeat_Z(ctx, instr, &emu_SCAS);
		} else {
			emu_repeatNZ(ctx, instr, &emu_SCAS);
		}
	} else 
	{
		u64_t di  = get_reg_addr(REG_CODE_DI);
		u64_t op0 = get_reggen(REG_CODE_AX, instr->opsize);
		u64_t op1 = read_mem_size(SREG_CODE_ES, di, instr->opsize);
		u64_t dif = op0 - op1;

		SET_FLAGS_OSZAPC_SUB(op0, op1, dif, instr->opsize);

		if (_DF != 0) {
			di -= instr->opsize;
		} else {
			di += instr->opsize;
		}
		set_reg_addr(REG_CODE_DI, di);

		/* zero extension of RDI */
		if (ctx->params->mode == EMU_MODE_32) RDI = EDI;
	}
}

void emu_CMPS(emu_context *ctx, struct INSTRUCTION *instr)
{
	if ( (instr->prefixes & INSTR_PREFIX_REP_MASK) && (ctx->in_repeat == 0) ) 
	{
		if (instr->prefixes & INSTR_PREFIX_REPZ) {
			emu_repeat_Z(ctx, instr, &emu_CMPS);
		} else {
			emu_repeatNZ(ctx, instr, &emu_CMPS);
		}
	} else
	{
		u64_t si  = get_reg_addr(REG_CODE_SI);
		u64_t di  = get_reg_addr(REG_CODE_DI);
		u64_t op0 = read_mem_size(SREG_CODE_DS, si, instr->opsize);
		u64_t op1 = read_mem_size(SREG_CODE_ES, di, instr->opsize);
		u64_t dif = op0 - op1;

		SET_FLAGS_OSZAPC_SUB(op0, op1, dif, instr->opsize);

		if (_DF != 0) {
			si -= instr->opsize; di -= instr->opsize;
		} else {
			si += instr->opsize; di += instr->opsize;
		}
		set_reg_addr(REG_CODE_SI, si);
		set_reg_addr(REG_CODE_DI, di);

		/* zero extension of RSI/RDI */
		if (ctx->params->mode == EMU_MODE_32) {
			RSI = ESI; RDI = EDI;
		}
	}
}

void emu_RETF(emu_context *ctx, struct INSTRUCTION *instr)
{
	emu_RETN(ctx, instr); popSEG(ctx, SREG_CODE_CS);
}

void emu_IRET(emu_context *ctx, struct INSTRUCTION *instr)
{
	emu_RETF(ctx, instr); emu_POPF(ctx, instr);
}

void emu_RETN(emu_context *ctx, struct INSTRUCTION *instr)
{
	switch (ctx->params->mode) {
		case EMU_MODE_16: EIP = pop16();  SP += instr->ops[0].value.imm.imm16; break;
		case EMU_MODE_32: EIP = pop32(); ESP += instr->ops[0].value.imm.imm16; break;
		case EMU_MODE_64: RIP = pop64(); RSP += instr->ops[0].value.imm.imm16; break;
	}
	ctx->jumped = 1;
}

void emu_ENTER(emu_context *ctx, struct INSTRUCTION *instr)
{
	u16_t imm16 = instr->ops[0].value.imm.imm16;
	u8_t  level = instr->ops[1].value.imm.imm8 & 0x1F;

	/* temporary fix for disasm bug */
	if (instr->flags & INSTR_FLAG_SF_PREFIXES) {
		instr->opsize = 2;
	} else {
		instr->opsize = 4;
	}
	switch (ctx->params->mode) {
		case EMU_MODE_16: EMU_DO_ENTER(SP,  BP,  level, imm16, 16, 2); break;
		case EMU_MODE_32: EMU_DO_ENTER(ESP, EBP, level, imm16, 32, instr->opsize); break;
		case EMU_MODE_64: EMU_DO_ENTER(RSP, RBP, level, imm16, 64, instr->opsize); break;
	}
}

void emu_LEAVE(emu_context *ctx, struct INSTRUCTION *instr)
{
	switch (ctx->params->mode) {
		case EMU_MODE_16: EMU_DO_LEAVE(SP,  BP,  16); break;
		case EMU_MODE_32: EMU_DO_LEAVE(ESP, EBP, 32); break;
		case EMU_MODE_64: EMU_DO_LEAVE(RSP, RBP, 64); break;
	}
}

void emu_HLT(emu_context *ctx, struct INSTRUCTION *instr)
{
	if (ctx->params->mode != EMU_MODE_16 && CPL != 0) {
		exception(EMU_EXCEPT_GP, 0);
	}
}

void emu_RDTSC(emu_context *ctx, struct INSTRUCTION *instr)
{
	if (ctx->regs_ctl.CR4.TSD != 0 || CPL == 0)
	{
		RAX = ctx->ticks & 0xFFFFFFFF;
		RDX = ctx->ticks >> 32;
	} else {
		exception(EMU_EXCEPT_GP, 0);
	}
}

void emu_RDTSCP(emu_context *ctx, struct INSTRUCTION *instr)
{
	emu_RDTSC(ctx, instr); RCX = ctx->regs_msr.tsc_aux;	
}

void emu_UD2(emu_context *ctx, struct INSTRUCTION *instr)
{
	exception(EMU_EXCEPT_UD, 0);
}

void emu_LOOP(emu_context *ctx, struct INSTRUCTION *instr)
{
	switch (ctx->params->mode) {
		case EMU_MODE_16: if (--CX  != 0) emu_JMP(ctx, instr); break;
		case EMU_MODE_32: if (--ECX != 0) emu_JMP(ctx, instr); break;
		case EMU_MODE_64: if (--RCX != 0) emu_JMP(ctx, instr); break;
	}
}

void emu_LOOPZ(emu_context *ctx, struct INSTRUCTION *instr)
{
	switch (ctx->params->mode) {
		case EMU_MODE_16: if (--CX  != 0 && _ZF != 0) emu_JMP(ctx, instr); break;
		case EMU_MODE_32: if (--ECX != 0 && _ZF != 0) emu_JMP(ctx, instr); break;
		case EMU_MODE_64: if (--RCX != 0 && _ZF != 0) emu_JMP(ctx, instr); break;
	}
}

void emu_LOOPNZ(emu_context *ctx, struct INSTRUCTION *instr)
{
	switch (ctx->params->mode) {
		case EMU_MODE_16: if (--CX  != 0 && _ZF == 0) emu_JMP(ctx, instr); break;
		case EMU_MODE_32: if (--ECX != 0 && _ZF == 0) emu_JMP(ctx, instr); break;
		case EMU_MODE_64: if (--RCX != 0 && _ZF == 0) emu_JMP(ctx, instr); break;
	}
}

void emu_XADD(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);
	u64_t sum = op0 + op1;

	save_operand(0, sum, 0);
	
	if ( (instr->ops[0].flags != instr->ops[1].flags) || (OP0_REG_CODE != OP1_REG_CODE) ) {
		save_operand(1, op0, 0);
	}
	SET_FLAGS_OSZAPC_ADD(op0, op1, sum, OP0_SIZE);
}

void emu_XCHG(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op0 = load_operand(0, 0);
	u64_t op1 = load_operand(1, 0);

	save_operand(0, op1, 0);
	save_operand(1, op0, 0);
}

void emu_MOVBE(emu_context *ctx, struct INSTRUCTION *instr)
{
	u64_t op = load_operand(1, 0);

	switch (OP0_SIZE) {
		case OPERAND_SIZE_16: op = BSWAP_16(op); break;
		case OPERAND_SIZE_32: op = BSWAP_32(op); break;
		case OPERAND_SIZE_64: op = BSWAP_64(op); break;
	}
	save_operand(0, op, 0);
}

void emu_XLAT(emu_context *ctx, struct INSTRUCTION *instr)
{
	AL = read_mem8(SREG_CODE_DS, (RBX + AL));
}