#include <stdio.h>
#include "mediana.h"
#include "x86_64_emu.h"
#include "emu_misc.h"

const u64_t sign_bits[] = { 0, 0x80, 0x8000, 0, 0x80000000, 0, 0, 0, 0x8000000000000000 };
const u64_t mask_bits[] = { 0, 0xFF, 0xFFFF, 0, 0xFFFFFFFF, 0, 0, 0, 0xFFFFFFFFFFFFFFFF };
const u64_t mode_mask[] = { 0, 0xFFFF, 0xFFFFFFFF, 0, 0xFFFFFFFFFFFFFFFF };

const u8_t parity_table[256] = {
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1
};

const u32_t reg8_offset[20] = {
	offsetof(emu_context, regs_gen.AL),   // REG_CODE_AX
	offsetof(emu_context, regs_gen.CL),   // REG_CODE_CX
	offsetof(emu_context, regs_gen.DL),   // REG_CODE_DX
	offsetof(emu_context, regs_gen.BL),   // REG_CODE_BX
	offsetof(emu_context, regs_gen.AH),   // REG_CODE_AH
	offsetof(emu_context, regs_gen.CH),   // REG_CODE_CH
	offsetof(emu_context, regs_gen.DH),   // REG_CODE_DH
	offsetof(emu_context, regs_gen.BH),   // REG_CODE_BH
	offsetof(emu_context, regs_gen.R8B),  // REG_CODE_R8
	offsetof(emu_context, regs_gen.R9B),  // REG_CODE_R9
	offsetof(emu_context, regs_gen.R10B), // REG_CODE_R10
	offsetof(emu_context, regs_gen.R11B), // REG_CODE_R11
	offsetof(emu_context, regs_gen.R12B), // REG_CODE_R12
	offsetof(emu_context, regs_gen.R13B), // REG_CODE_R13
	offsetof(emu_context, regs_gen.R14B), // REG_CODE_R14
	offsetof(emu_context, regs_gen.R15B), // REG_CODE_R15
	offsetof(emu_context, regs_gen.SPL),  // REG_CODE_SPL
	offsetof(emu_context, regs_gen.BPL),  // REG_CODE_BPL
	offsetof(emu_context, regs_gen.SIL),  // REG_CODE_SIL
	offsetof(emu_context, regs_gen.DIL)   // REG_CODE_DIL
};

u64_t emu_get_address(emu_context *ctx, struct INSTRUCTION *instr, int op_n, int *sreg)
{
	u64_t offset = 0;

	if (instr->ops[op_n].value.addr.mod & ADDR_MOD_BASE) {
		offset += get_reg_addr(instr->ops[op_n].value.addr.base);
	}
	if (instr->ops[op_n].value.addr.mod & ADDR_MOD_IDX) 
	{
		offset += get_reg_addr(instr->ops[op_n].value.addr.index) * 
			      instr->ops[op_n].value.addr.scale;
	}
	if (instr->ops[op_n].value.addr.mod & ADDR_MOD_DISP) {
		offset += instr->disp.value.d64;
	}
	if (sreg != NULL)
	{
		switch (instr->prefixes & INSTR_PREFIX_SEG_MASK) {
			case INSTR_PREFIX_CS: *sreg = SREG_CODE_CS; break;
			case INSTR_PREFIX_DS: *sreg = SREG_CODE_DS; break;
			case INSTR_PREFIX_ES: *sreg = SREG_CODE_ES; break;
			case INSTR_PREFIX_SS: *sreg = SREG_CODE_SS; break;
			case INSTR_PREFIX_FS: *sreg = SREG_CODE_FS; break;
			case INSTR_PREFIX_GS: *sreg = SREG_CODE_GS; break;
			default: *sreg = SREG_CODE_DS;
		}
	}
	return offset & mask_bits[instr->addrsize];
}

static u64_t emu_read_mem(emu_context *ctx, struct INSTRUCTION *instr, int sreg, u64_t offset, int size)
{
	u16_t seg = get_regseg(sreg);
	u64_t off = instr != NULL ? MASK_ADDR(offset) : MASK_MODE(offset);

	switch (size) {
		case OPERAND_SIZE_8:  return ctx->params->do_loadmem_8(ctx, seg, off);
		case OPERAND_SIZE_16: return ctx->params->do_loadmem16(ctx, seg, off);
		case OPERAND_SIZE_32: return ctx->params->do_loadmem32(ctx, seg, off);
		case OPERAND_SIZE_64: return ctx->params->do_loadmem64(ctx, seg, off);
	}
	exception(EMU_EXCEPT_UD, 0);
}

static void emu_write_mem(emu_context *ctx, struct INSTRUCTION *instr, int sreg, u64_t offset, u64_t data, int size)
{
	u16_t seg = get_regseg(sreg);
	u64_t off = instr != NULL ? MASK_ADDR(offset) : MASK_MODE(offset);

	switch (size) {
		case OPERAND_SIZE_8:  ctx->params->do_savemem_8(ctx, seg, off, u8t (data)); return;
		case OPERAND_SIZE_16: ctx->params->do_savemem16(ctx, seg, off, u16t(data)); return;
		case OPERAND_SIZE_32: ctx->params->do_savemem32(ctx, seg, off, u32t(data)); return;
		case OPERAND_SIZE_64: ctx->params->do_savemem64(ctx, seg, off, u64t(data)); return;
	}
	exception(EMU_EXCEPT_UD, 0);
}

u64_t emu_load_operand(emu_context *ctx, struct INSTRUCTION *instr, int op_n, u64_t mem_off)
{
	if (instr->ops[op_n].flags & OPERAND_TYPE_REG)
	{
		switch (instr->ops[op_n].size) {
			case OPERAND_SIZE_8:  return get_reggen8(instr->ops[op_n].value.reg.code);
			case OPERAND_SIZE_16: return get_reggen16(instr->ops[op_n].value.reg.code);
			case OPERAND_SIZE_32: return get_reggen32(instr->ops[op_n].value.reg.code);
			case OPERAND_SIZE_64: return get_reggen64(instr->ops[op_n].value.reg.code);
		}
	}
	if (instr->ops[op_n].flags & OPERAND_TYPE_MEM)
	{
		int   sreg;
		u64_t offs = emu_get_address(ctx, instr, op_n, &sreg) + mem_off;
		
		return emu_read_mem(ctx, instr, sreg, offs, instr->ops[op_n].size);
	}
	if (instr->ops[op_n].flags & OPERAND_TYPE_IMM) {
		return load_imm_op(op_n);
	}
	exception(EMU_EXCEPT_UD, 0);
}

void emu_save_operand(emu_context *ctx, struct INSTRUCTION *instr, u64_t data, int op_n, u64_t mem_off)
{
	if (instr->ops[op_n].flags & OPERAND_TYPE_REG)
	{
		switch (instr->ops[op_n].size) {
			case OPERAND_SIZE_8:  set_reggen8(instr->ops[op_n].value.reg.code, data); return;
			case OPERAND_SIZE_16: set_reggen16(instr->ops[op_n].value.reg.code, data); return;
			case OPERAND_SIZE_32: set_reggen32(instr->ops[op_n].value.reg.code, data); return;
			case OPERAND_SIZE_64: set_reggen64(instr->ops[op_n].value.reg.code, data); return;
		}
	}
	if (instr->ops[op_n].flags & OPERAND_TYPE_MEM) 
	{
		int   sreg;
		u64_t offs = emu_get_address(ctx, instr, op_n, &sreg) + mem_off;
		
		emu_write_mem(ctx, instr, sreg, offs, data, instr->ops[op_n].size);
		return;	
	}
	exception(EMU_EXCEPT_UD, 0);
}


void emu_push(emu_context *ctx, u64_t data, u16_t opsize)
{
	emu_write_mem(
		ctx, NULL, SREG_CODE_SS, ctx->regs_gen.RSP - opsize, data, opsize);

	switch (ctx->params->mode) {
		case EMU_MODE_16: ctx->regs_gen.SP  -= opsize; break;
		case EMU_MODE_32: ctx->regs_gen.ESP -= opsize; break;
		case EMU_MODE_64: ctx->regs_gen.RSP -= opsize; break;
	}
}

u64_t emu_pop(emu_context *ctx, u16_t opsize)
{
	u64_t data = emu_read_mem(
		ctx, NULL, SREG_CODE_SS, ctx->regs_gen.RSP, opsize);

	switch (ctx->params->mode) {
		case EMU_MODE_16: ctx->regs_gen.SP  += opsize; break;
		case EMU_MODE_32: ctx->regs_gen.ESP += opsize; break;
		case EMU_MODE_64: ctx->regs_gen.RSP += opsize; break;
	}
	return data;
}