#include "x86_64_emu.h"
#include "mediana.h"
#include "emu_defines.h"
#include "emu_general.h"
#include "emu_cpuid.h"
#ifdef EMU_DBG_DUMP
#include <stdio.h>
#endif

static int emu_execute(emu_context *ctx, struct INSTRUCTION *instr)
{
	emu_params *params = ctx->params;
	int         status;

	ctx->ticks++;

	if (params->custom_ops[instr->id].pre_op != NULL) 
	{
		if ( (status = params->custom_ops[instr->id].pre_op(ctx, instr)) != EMU_STATUS_OK) {
			return status;
		}
	}
	if (params->custom_ops[instr->id].new_op != NULL) 
	{
		if ( (status = params->custom_ops[instr->id].new_op(ctx, instr)) != EMU_STATUS_OK) {
			return status;
		}
	} else
	{
		switch (instr->id) {
			/* decimal */
			case ID_AAA: emu_AAA(ctx, instr); break;
			case ID_AAD: emu_AAD(ctx, instr); break;
			case ID_AAM: emu_AAM(ctx, instr); break;
			case ID_AAS: emu_AAS(ctx, instr); break;
			case ID_DAA: emu_DAA(ctx, instr); break;
			case ID_DAS: emu_DAS(ctx, instr); break;
			/* generic */
			case ID_MOV:       emu_MOV(ctx, instr);       break;
			case ID_LEA:       emu_LEA(ctx, instr);       break;
			case ID_CMP:       emu_CMP(ctx, instr);       break;
			case ID_CMPXCHG:   emu_CMPXCHG(ctx, instr);   break;
			case ID_CMPXCHG8B: emu_CMPXCHG8B(ctx, instr); break;
			case ID_MOVSX:     emu_MOVSX(ctx, instr);     break;
			case ID_MOVSXD:    emu_MOVSXD(ctx, instr);    break;
			case ID_MOVZX:     emu_MOVZX(ctx, instr);     break;
			case ID_XLAT:      emu_XLAT(ctx, instr);      break;
			/* arithmetical */
			case ID_ADD:  emu_ADD(ctx, instr);  break;
			case ID_ADC:  emu_ADC(ctx, instr);  break;
			case ID_SUB:  emu_SUB(ctx, instr);  break;
			case ID_SBB:  emu_SBB(ctx, instr);  break;
			case ID_INC:  emu_INC(ctx, instr);  break;
			case ID_DEC:  emu_DEC(ctx, instr);  break;
			case ID_DIV:  emu_DIV(ctx, instr);  break;
			case ID_IDIV: emu_IDIV(ctx, instr); break;
			case ID_MUL:  emu_MUL(ctx, instr);  break;
			case ID_IMUL: emu_IMUL(ctx, instr); break;
			case ID_CBW:  emu_CBW(ctx, instr);  break;
			case ID_CWD:  emu_CWD(ctx, instr);  break;
			case ID_NEG:  emu_NEG(ctx, instr);  break;
			case ID_XADD: emu_XADD(ctx, instr); break;
			/* logical */
			case ID_AND:  emu_AND(ctx, instr);  break;
			case ID_OR:   emu_OR(ctx, instr);   break;
			case ID_NOT:  emu_NOT(ctx, instr);  break;
			case ID_XOR:  emu_XOR(ctx, instr);  break;
			case ID_TEST: emu_TEST(ctx, instr); break;

			/* shift instructons */
			case ID_ROL:  emu_ROL(ctx, instr);  break;
			case ID_ROR:  emu_ROR(ctx, instr);  break;
			case ID_RCL:  emu_RCL(ctx, instr);  break;
			case ID_RCR:  emu_RCR(ctx, instr);  break;
			case ID_SAL:  emu_SHL(ctx, instr);  break;
			case ID_SAR:  emu_SAR(ctx, instr);  break;
			case ID_SHL:  emu_SHL(ctx, instr);  break;
			case ID_SHR:  emu_SHR(ctx, instr);  break;
			case ID_SHLD: emu_SHLD(ctx, instr); break;
			case ID_SHRD: emu_SHRD(ctx, instr); break;

			/* flags control */
			case ID_CLC:   emu_CLC(ctx, instr);   break;
			case ID_STC:   emu_STC(ctx, instr);   break;
			case ID_CLD:   emu_CLD(ctx, instr);   break;
			case ID_STD:   emu_STD(ctx, instr);   break;
			case ID_CLI:   emu_CLI(ctx, instr);   break;
			case ID_STI:   emu_STI(ctx, instr);   break;
			case ID_CMC:   emu_CMC(ctx, instr);   break;
			case ID_SAHF:  emu_SAHF(ctx, instr);  break;
			case ID_LAHF:  emu_LAHF(ctx, instr);  break;
			case ID_PUSHF: emu_PUSHF(ctx, instr); break;
			case ID_POPF:  emu_POPF(ctx, instr);  break;
			case ID_SALC:  emu_SALC(ctx, instr);  break;

			/* bit operations */
			case ID_BSF:   emu_BSF(ctx, instr);   break;
			case ID_BSR:   emu_BSR(ctx, instr);   break;
			case ID_BT:    emu_BT(ctx, instr);    break;
			case ID_BTC:   emu_BTC(ctx, instr);   break;
			case ID_BTR:   emu_BTR(ctx, instr);   break;
			case ID_BTS:   emu_BTS(ctx, instr);   break;
			case ID_BSWAP: emu_BSWAP(ctx, instr); break;
			case ID_MOVBE: emu_MOVBE(ctx, instr); break;

			/* misc */
			case ID_ARPL:  emu_ARPL(ctx, instr);  break;
			case ID_BOUND: emu_BOUND(ctx, instr); break;
			case ID_NOP:                          break;
			case ID_UD:    emu_UD2(ctx, instr);   break;
			case ID_UD2:   emu_UD2(ctx, instr);   break;
			case ID_XCHG:  emu_XCHG(ctx, instr);  break;

			/* system */
			case ID_CLTS:   emu_CLTS(ctx, instr);   break;
			case ID_HLT:    emu_HLT(ctx, instr);    break;
			case ID_INVLPG: emu_HLT(ctx, instr);    break;
			case ID_INVD:   emu_HLT(ctx, instr);    break;
			case ID_WBINVD: emu_HLT(ctx, instr);    break;
			case ID_RDTSC:  emu_RDTSC(ctx, instr);  break;
			case ID_RDTSCP: emu_RDTSCP(ctx, instr); break;
			case ID_CPUID:  emu_CPUID(ctx, instr);  break;

			/* string */
			case ID_LODS: emu_LODS(ctx, instr); break;
			case ID_STOS: emu_STOS(ctx, instr); break;
			case ID_MOVS: emu_MOVS(ctx, instr); break;
			case ID_SCAS: emu_SCAS(ctx, instr); break;
			case ID_CMPS: emu_CMPS(ctx, instr); break;

			/* stack */
			case ID_PUSH:  emu_PUSH(ctx, instr);  break;
			case ID_PUSHA: emu_PUSHA(ctx, instr); break;
			case ID_POP:   emu_POP(ctx, instr);   break;
			case ID_POPA:  emu_POPA(ctx, instr);  break;
			case ID_ENTER: emu_ENTER(ctx, instr); break;
			case ID_LEAVE: emu_LEAVE(ctx, instr); break;

			/* some branches */
			case ID_JMP:    emu_JMP(ctx, instr);    break;
			case ID_JMPF:   emu_JMPF(ctx, instr);   break;
			case ID_CALL:   emu_CALL(ctx, instr);   break;
			case ID_CALLF:  emu_CALLF(ctx, instr);  break;
			case ID_RETN:   emu_RETN(ctx, instr);   break;
			case ID_RETF:   emu_RETF(ctx, instr);   break;
			case ID_IRET:   emu_IRET(ctx, instr);   break;
			case ID_JCXZ:   emu_JCXZ(ctx, instr);   break;
			case ID_LOOP:   emu_LOOP(ctx, instr);   break;
			case ID_LOOPZ:  emu_LOOPZ(ctx, instr);  break;
			case ID_LOOPNZ: emu_LOOPNZ(ctx, instr); break;

			/* CMOVcc */
			case ID_CMOVA:  emu_CMOVcc(ctx, instr, (_CF == 0 && _ZF == 0)  ); break;
			case ID_CMOVAE: emu_CMOVcc(ctx, instr, (_CF == 0)              ); break;
			case ID_CMOVB:  emu_CMOVcc(ctx, instr, (_CF != 0)              ); break;
			case ID_CMOVBE: emu_CMOVcc(ctx, instr, (_CF != 0 || _ZF != 0)  ); break;
			case ID_CMOVG:  emu_CMOVcc(ctx, instr, (_ZF == 0 && _SF == _OF)); break;
			case ID_CMOVGE: emu_CMOVcc(ctx, instr, (_SF == _OF)            ); break;
			case ID_CMOVL:  emu_CMOVcc(ctx, instr, (_SF != _OF)            ); break;
			case ID_CMOVLE: emu_CMOVcc(ctx, instr, (_ZF != 0 || _SF != _OF)); break;
			case ID_CMOVNO: emu_CMOVcc(ctx, instr, (_OF == 0)              ); break;
			case ID_CMOVNP: emu_CMOVcc(ctx, instr, (_PF == 0)              ); break;
			case ID_CMOVNS: emu_CMOVcc(ctx, instr, (_SF == 0)              ); break;
			case ID_CMOVNZ: emu_CMOVcc(ctx, instr, (_ZF == 0)              ); break;
			case ID_CMOVO:  emu_CMOVcc(ctx, instr, (_OF != 0)              ); break;
			case ID_CMOVP:  emu_CMOVcc(ctx, instr, (_PF != 0)              ); break;
			case ID_CMOVS:  emu_CMOVcc(ctx, instr, (_SF != 0)              ); break;
			case ID_CMOVZ:  emu_CMOVcc(ctx, instr, (_ZF != 0)              ); break;
			/* Jcc */
			case ID_JA:  emu_Jcc(ctx, instr, (_CF == 0 && _ZF == 0)  ); break;
			case ID_JAE: emu_Jcc(ctx, instr, (_CF == 0)              ); break;
			case ID_JB:  emu_Jcc(ctx, instr, (_CF != 0)              ); break;
			case ID_JBE: emu_Jcc(ctx, instr, (_CF != 0 || _ZF != 0)  ); break;
			case ID_JG:  emu_Jcc(ctx, instr, (_ZF == 0 && _SF == _OF)); break;
			case ID_JGE: emu_Jcc(ctx, instr, (_SF == _OF)            ); break;
			case ID_JL:  emu_Jcc(ctx, instr, (_SF != _OF)            ); break;
			case ID_JLE: emu_Jcc(ctx, instr, (_ZF != 0 || _SF != _OF)); break;
			case ID_JNO: emu_Jcc(ctx, instr, (_OF == 0)              ); break;
			case ID_JNP: emu_Jcc(ctx, instr, (_PF == 0)              ); break;
			case ID_JNS: emu_Jcc(ctx, instr, (_SF == 0)              ); break;
			case ID_JNZ: emu_Jcc(ctx, instr, (_ZF == 0)              ); break;
			case ID_JO:  emu_Jcc(ctx, instr, (_OF != 0)              ); break;
			case ID_JP:  emu_Jcc(ctx, instr, (_PF != 0)              ); break;
			case ID_JS:  emu_Jcc(ctx, instr, (_SF != 0)              ); break;
			case ID_JZ:  emu_Jcc(ctx, instr, (_ZF != 0)              ); break;
			/* SETcc */
			case ID_SETA:  emu_SETcc(ctx, instr, (_CF == 0 && _ZF == 0)  ); break;
			case ID_SETAE: emu_SETcc(ctx, instr, (_CF == 0)              ); break;
			case ID_SETB:  emu_SETcc(ctx, instr, (_CF != 0)              ); break;
			case ID_SETBE: emu_SETcc(ctx, instr, (_CF != 0 || _ZF != 0)  ); break;
			case ID_SETG:  emu_SETcc(ctx, instr, (_ZF == 0 && _SF == _OF)); break;
			case ID_SETGE: emu_SETcc(ctx, instr, (_SF == _OF)            ); break;
			case ID_SETL:  emu_SETcc(ctx, instr, (_SF != _OF)            ); break;
			case ID_SETLE: emu_SETcc(ctx, instr, (_ZF != 0 || _SF != _OF)); break;
			case ID_SETNO: emu_SETcc(ctx, instr, (_OF == 0)              ); break;
			case ID_SETNP: emu_SETcc(ctx, instr, (_PF == 0)              ); break;
			case ID_SETNS: emu_SETcc(ctx, instr, (_SF == 0)              ); break;
			case ID_SETNZ: emu_SETcc(ctx, instr, (_ZF == 0)              ); break;
			case ID_SETO:  emu_SETcc(ctx, instr, (_OF != 0)              ); break;
			case ID_SETP:  emu_SETcc(ctx, instr, (_PF != 0)              ); break;
			case ID_SETS:  emu_SETcc(ctx, instr, (_SF != 0)              ); break;
			case ID_SETZ:  emu_SETcc(ctx, instr, (_ZF != 0)              ); break;

			default: exception(EMU_EXCEPT_UD, 0);
		}
	}
	if (params->custom_ops[instr->id].post_op != NULL) 
	{
		if ( (status = params->custom_ops[instr->id].post_op(ctx, instr)) != EMU_STATUS_OK) {
			return status;
		}
	}
	if (ctx->regs_gen.flags._tf != 0) {
		exception(EMU_EXCEPT_DB, 0);
	}
	return EMU_STATUS_OK;
}


int emu_execute_code(emu_context *ctx, emu_params *params)
{
	struct INSTRUCTION   instr;
	struct DISASM_PARAMS dasmp;
	unsigned char        buff[16], *p_code;
	int                  status;
	u64_t                initial_ticks;

	if (setjmp(params->saved_env) != 0)
	{
		if ( (params->on_interrupt == NULL) || 
			 ((status = params->on_interrupt(ctx)) != EMU_STATUS_OK) )
		{
			return status;
		}
		ctx->jumped = 1; goto exec_loop;
	}
	dasmp.sf_prefixes     = NULL;
	dasmp.sf_prefixes_len = 0;
	dasmp.arch    = ARCH_COMMON | ARCH_INTEL;
	dasmp.mode    = params->mode;
	dasmp.options = 0;
	dasmp.base    = 0;
	ctx->jumped   = 1;
	ctx->params   = params;
	initial_ticks = ctx->ticks;

	for (;;)
	{
exec_loop:
		/* fetch instruction */
		if (params->on_fetchcode != NULL) 
		{
			if ( (status = params->on_fetchcode(ctx, buff)) != EMU_STATUS_OK ) {
				return status;
			}
			p_code = buff; ctx->jumped = 0;
		} else 
		{
			if ( (ctx->jumped != 0) &&
				 ((p_code = params->on_resolveip(ctx)) == NULL) )
			{
				return EMU_STATUS_FETCH_ERROR;
			} else {
				ctx->jumped = 0;
			}
		}
		/* disassemble opcode */
		if ( (status = medi_disassemble(p_code, 16, &instr, &dasmp)) != DASM_ERR_OK ) {
			if (status == DASM_ERR_INTERNAL) return EMU_STATUS_INTERNAL_ERROR;
			exception(EMU_EXCEPT_UD, 0);
		}
#ifdef EMU_DBG_DUMP
		{
			char buff[64] = {0};

			medi_dump(&instr, buff, sizeof(buff), NULL);
			printf("%s\n", buff);
		}
#endif
		/* execute instruction */
		if ( (status = emu_execute(ctx, &instr)) != EMU_STATUS_OK ) {
			return status;
		}
		/* update RIP and code pointer */
		if (ctx->jumped != 0) {
			RIP = RIP & mode_mask[params->mode];
			if (params->flags & EMU_FLAG_STOP_ON_BRANCH) return EMU_STATUS_BRANCH;			
		} else {
			RIP     = (RIP + instr.length) & mode_mask[params->mode];
			p_code += instr.length;
		}
		/* check execution limit */
		if ( (params->flags & EMU_FLAG_SINGLE_INSTR) ||
			 ((params->flags & EMU_FLAG_LIMIT) && (ctx->ticks - initial_ticks >= params->limit)) )
		{
			return EMU_STATUS_OK;
		}
	}
}
