#ifndef _X86_64_EMU_H_
#define _X86_64_EMU_H_

#include <setjmp.h>
#include "mediana.h"

typedef unsigned long long u64_t;
typedef unsigned long      u32_t;
typedef unsigned short     u16_t;
typedef unsigned char      u8_t;
typedef long long          s64_t;
typedef long               s32_t;
typedef short              s16_t;
typedef char               s8_t;

typedef struct { u64_t lo; u64_t hi; } u128_t;
typedef struct { u64_t lo; s64_t hi; } s128_t;


#define u64t(_x) (u64_t)(_x)
#define u32t(_x) (u32_t)(_x)
#define u16t(_x) (u16_t)(_x)
#define u8t(_x)  (u8_t)(_x)

#define s64t(_x) (s64_t)(_x)
#define s32t(_x) (s32_t)(_x)
#define s16t(_x) (s16_t)(_x)
#define s8t(_x)  (s8_t)(_x)

/* emulation flags */
#define EMU_FLAG_SINGLE_INSTR      0x01 /* execute single instruction                          */
#define EMU_FLAG_STOP_ON_BRANCH    0x02 /* stop execution after first branch                   */
#define EMU_FLAG_SLOW_REP          0x04 /* treat repeated instruction as multiple instructions */
#define EMU_FLAG_LIMIT             0x08 /* limit number of executed instructions               */

/* emulation modes */
#define EMU_MODE_16 DISASSEMBLE_MODE_16 /* 16bit real mode      */
#define EMU_MODE_32 DISASSEMBLE_MODE_32 /* 32bit protected mode */
#define EMU_MODE_64 DISASSEMBLE_MODE_64 /* 64bit long mode      */

#define EMU_STATUS_OK             0
#define EMU_STATUS_INTERRUPT      1
#define EMU_STATUS_FETCH_ERROR    2
#define EMU_STATUS_INTERNAL_ERROR 3
#define EMU_STATUS_BRANCH         4
#define EMU_ABORT_OPERATION       5

/* callback types */
typedef int   (*emu_fetchopcode)(struct _emu_context *ctx, u8_t buff[16]);
typedef void* (*emu_resolveip)  (struct _emu_context *ctx);
typedef int   (*emu_intcallback)(struct _emu_context *ctx);
typedef int   (*emu_loadspecreg)(struct _emu_context *ctx, u8_t type, u8_t code, u64_t data);

typedef u8_t  (*emu_loadmem_8)(struct _emu_context *ctx, u16_t seg, u64_t offset);
typedef u16_t (*emu_loadmem16)(struct _emu_context *ctx, u16_t seg, u64_t offset);
typedef u32_t (*emu_loadmem32)(struct _emu_context *ctx, u16_t seg, u64_t offset);
typedef u64_t (*emu_loadmem64)(struct _emu_context *ctx, u16_t seg, u64_t offset);
typedef void  (*emu_savemem_8)(struct _emu_context *ctx, u16_t seg, u64_t offset, u8_t  data);
typedef void  (*emu_savemem16)(struct _emu_context *ctx, u16_t seg, u64_t offset, u16_t data);
typedef void  (*emu_savemem32)(struct _emu_context *ctx, u16_t seg, u64_t offset, u32_t data);
typedef void  (*emu_savemem64)(struct _emu_context *ctx, u16_t seg, u64_t offset, u64_t data);

typedef int   (*emu_ophandler)(struct _emu_context *ctx, struct INSTRUCTION *instr);

typedef struct _emu_params {
	u32_t flags;   /* execution flags, see EMU_FLAG_xxx                           */
	u64_t limit;   /* maximum instructions count, valid if EMU_FLAG_LIMIT present */
	u8_t  int_num; /* last interrupt number    */
	u64_t int_arg; /* last interrupt parameter */
	int   mode;

	/* callbacks */
	emu_intcallback on_interrupt; /* interrupt callback                   */
	emu_fetchopcode on_fetchcode; /* fetch opcode callback                */
	emu_resolveip   on_resolveip; /* resolve instruction pointer callback */
	emu_loadspecreg do_loadsreg;  /* load special register                */

	/* memory callbacks */
	emu_loadmem_8   do_loadmem_8;
	emu_loadmem16   do_loadmem16;
	emu_loadmem32   do_loadmem32;
	emu_loadmem64   do_loadmem64;
	emu_savemem_8   do_savemem_8;
	emu_savemem16   do_savemem16;
	emu_savemem32   do_savemem32;
	emu_savemem64   do_savemem64;

	/* instruction callbacks */
	struct {
		emu_ophandler pre_op;
		emu_ophandler new_op;
		emu_ophandler post_op;

	} custom_ops[ID_COUNT];

	/* for internal usage, DON't use */
	jmp_buf saved_env;
} emu_params;

typedef struct _emu_context {
	union 
	{
		union {
			u64_t reg64; u32_t reg32; u16_t reg16; struct { u8_t reg8l; u8_t reg8h; };
		} regs[22];
		
		struct {
			union { u64_t RAX; u32_t EAX; u16_t AX; struct { u8_t AL; u8_t AH; }; };
			union { u64_t RCX; u32_t ECX; u16_t CX; struct { u8_t CL; u8_t CH; }; };
			union { u64_t RDX; u32_t EDX; u16_t DX; struct { u8_t DL; u8_t DH; }; };
			union { u64_t RBX; u32_t EBX; u16_t BX; struct { u8_t BL; u8_t BH; }; };
			union { u64_t RSP; u32_t ESP; u16_t SP;  u8_t SPL; };
			union { u64_t RBP; u32_t EBP; u16_t BP;  u8_t BPL; };
			union { u64_t RSI; u32_t ESI; u16_t SI;  u8_t SIL; };
			union { u64_t RDI; u32_t EDI; u16_t DI;  u8_t DIL; };
			/* R8 - R15 */
			union { u64_t R8;  u32_t R8D;  u16_t R8W;  u8_t R8B;  };
			union { u64_t R9;  u32_t R9D;  u16_t R9W;  u8_t R9B;  };
			union { u64_t R10; u32_t R10D; u16_t R10W; u8_t R10B; };
			union { u64_t R11; u32_t R11D; u16_t R11W; u8_t R11B; };
			union { u64_t R12; u32_t R12D; u16_t R12W; u8_t R12B; };
			union { u64_t R13; u32_t R13D; u16_t R13W; u8_t R13B; };
			union { u64_t R14; u32_t R14D; u16_t R14W; u8_t R14B; };
			union { u64_t R15; u32_t R15D; u16_t R15W; u8_t R15B; };
			/* */
			u64_t unused[4];
			/* RIP, RFL */
			union { u64_t RIP; u32_t EIP; u16_t IP; };
			union 
			{ 
				u64_t RFL; u32_t EFL; u16_t FL; 

				struct {
					u16_t _cf:1;
					u16_t _unk0:1; /* always set to 1 */
					u16_t _pf:1;
					u16_t _unk1:1; /* always set to 0 */
					u16_t _af:1;
					u16_t _unk2:1; /* always set to 0 */
					u16_t _zf:1;
					u16_t _sf:1;
					u16_t _tf:1;
					u16_t _if:1;
					u16_t _df:1;
					u16_t _of:1;
					u16_t _iopl:2;
					u16_t _nt:1;
					u16_t _unk3:1; /* always set to 0 */
					u16_t _rf:1;
					u16_t _vm:1;
					u16_t _ac:1;
					u16_t _vif:1;
					u16_t _vip:1;
					u16_t _id:1;
					u16_t _unk4:10; /* always set to 0 */
				} flags;
			};
		};
	} regs_gen;

	union  {
		u16_t regs[6];
		struct { u16_t CS; u16_t DS; u16_t ES; u16_t SS; u16_t FS; u16_t GS; };
	} regs_seg;
	
	struct 
	{
		union {
			u64_t reg;

			struct { 
				u16_t PE:1; 
				u16_t MP:1; 
				u16_t EM:1; 
				u16_t TS:1; 
				u16_t ET:1;
				u16_t NE:1; 
				u16_t unused0:10; 
				u16_t WP:1; 
				u16_t unused1:1; 
				u16_t AM:1;
				u16_t unused2:10; 
				u16_t NW:1; 
				u16_t CD:1; 
				u16_t PG:1; 
			};
		} CR0;

		union { u64_t reg; } CR1;
		union { u64_t reg; } CR2;

		union { 
			u64_t reg; 
			
			struct { 
				u16_t unused0:3; 
				u16_t PWT:1; 
				u16_t unused1:7; 
				u64_t base:53; 
			};
		} CR3;

		union {
			u64_t reg; 

			struct {
				u16_t VME:1;
				u16_t PVI:1;
				u16_t TSD:1;
				u16_t DE:1;
				u16_t PSE:1;
				u16_t PAE:1;
				u16_t MCE:1;
				u16_t PGE:1;
				u16_t PCE:1;
				u16_t OSFXSR:1;
				u16_t OSXMMEXCPT:1;
				u16_t unused0:2;
				u16_t VMXE:1;
				u16_t SMXE:1;
				u16_t unused1:2;
				u16_t PCIDE:1;
				u16_t OSXSAVE:1;
			};
		} CR4;
		
	} regs_ctl;

	struct {
		u64_t tsc_aux;
	} regs_msr;

	/* for internal use */
	u64_t       ticks;
	int         jumped;
	int         in_repeat;
	emu_params *params;
} emu_context;



#define	OPERAND_SIZE_8    0x0001
#define	OPERAND_SIZE_16   0x0002
#define	OPERAND_SIZE_32   0x0004
#define	OPERAND_SIZE_48   0x0006
#define	OPERAND_SIZE_64   0x0008
#define	OPERAND_SIZE_80   0x000A
#define	OPERAND_SIZE_128  0x0010
#define OPERAND_SIZE_14   0x000E
#define OPERAND_SIZE_28   0x001C
#define OPERAND_SIZE_94   0x005E
#define OPERAND_SIZE_108  0x006C
#define OPERAND_SIZE_512  0x0200

#define EMU_EXCEPT_DE  0
#define EMU_EXCEPT_DB  1
#define EMU_EXCEPT_BP  3
#define EMU_EXCEPT_OF  4
#define EMU_EXCEPT_BR  5
#define EMU_EXCEPT_UD  6
#define EMU_EXCEPT_DF  8
#define EMU_EXCEPT_TS  10
#define EMU_EXCEPT_SS  12
#define EMU_EXCEPT_GP  13
#define EMU_EXCEPT_PF  14
#define EMU_EXCEPT_MC  18

int emu_execute_code(emu_context *ctx, emu_params *params);

#endif